package com.sun.jmx.mbeanserver;

import java.util.Collections;
import java.util.WeakHashMap;
import java.util.Locale;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.lang.ref.SoftReference;
import java.util.Map;
import java.security.AccessController;
import com.sun.jmx.remote.util.EnvHelp;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Array;
import javax.management.AttributeNotFoundException;
import javax.management.openmbean.CompositeData;
import java.util.Arrays;
import java.lang.reflect.Proxy;
import java.lang.reflect.Method;
import java.lang.reflect.UndeclaredThrowableException;
import sun.reflect.misc.MethodUtil;
import javax.management.DescriptorKey;
import java.util.HashMap;
import java.lang.annotation.Annotation;
import javax.management.ImmutableDescriptor;
import javax.management.Descriptor;
import java.lang.reflect.AnnotatedElement;
import sun.reflect.misc.ReflectUtil;
import javax.management.MBeanInfo;
import javax.management.NotCompliantMBeanException;
import java.lang.reflect.Modifier;
import javax.management.DynamicMBean;

public class Introspector
{
    public static final boolean ALLOW_NONPUBLIC_MBEAN;
    
    private Introspector() {
    }
    
    public static final boolean isDynamic(Class<?> clazz) {
        return DynamicMBean.class.isAssignableFrom(clazz);
    }
    
    public static void testCreation(Class<?> clazz) throws NotCompliantMBeanException {
        int modifiers = clazz.getModifiers();
        if (Modifier.isAbstract(modifiers) || Modifier.isInterface(modifiers)) {
            throw new NotCompliantMBeanException("MBean class must be concrete");
        }
        if (clazz.getConstructors().length == 0) {
            throw new NotCompliantMBeanException("MBean class must have public constructor");
        }
        return;
    }
    
    public static void checkCompliance(Class<?> clazz) throws NotCompliantMBeanException {
        if (DynamicMBean.class.isAssignableFrom(clazz)) {
            return;
        }
        try {
            getStandardMBeanInterface(clazz);
            return;
        }
        catch (NotCompliantMBeanException ex) {
            try {
                getMXBeanInterface(clazz);
                return;
            }
            catch (NotCompliantMBeanException ex2) {
                throw new NotCompliantMBeanException("MBean class " + clazz.getName() + " does not implement DynamicMBean, and neither follows the Standard MBean conventions (" + ex.toString() + ") nor the MXBean conventions (" + ex2.toString() + ")");
            }
        }
    }
    
    public static <T> DynamicMBean makeDynamicMBean(T t) throws NotCompliantMBeanException {
        if ((t instanceof DynamicMBean)) {
            return (DynamicMBean)t;
        }
        Class<Object> class1 = t.getClass();
        Class<Object> clazz = null;
        try {
            clazz = (Class)Util.cast(getStandardMBeanInterface(class1));
        }
        catch (NotCompliantMBeanException ex) {}
        if (clazz != null) {
            return new StandardMBeanSupport(t, clazz);
        }
        try {
            clazz = (Class)Util.cast(getMXBeanInterface(class1));
        }
        catch (NotCompliantMBeanException ex2) {}
        if (clazz != null) {
            return new MXBeanSupport(t, clazz);
        }
        checkCompliance(class1);
        throw new NotCompliantMBeanException("Not compliant");
    }
    
    public static MBeanInfo testCompliance(Class<?> clazz) throws NotCompliantMBeanException {
        if (isDynamic(clazz)) {
            return null;
        }
        return testCompliance(clazz, null);
    }
    
    public static void testComplianceMXBeanInterface(Class<?> clazz) throws NotCompliantMBeanException {
        MXBeanIntrospector.getInstance().getAnalyzer(clazz);
    }
    
    public static void testComplianceMBeanInterface(Class<?> clazz) throws NotCompliantMBeanException {
        StandardMBeanIntrospector.getInstance().getAnalyzer(clazz);
    }
    
    public static synchronized MBeanInfo testCompliance(Class<?> clazz, Class<?> standardMBeanInterface) throws NotCompliantMBeanException {
        if (standardMBeanInterface == null) {
            standardMBeanInterface = getStandardMBeanInterface(clazz);
        }
        ReflectUtil.checkPackageAccess(standardMBeanInterface);
        return getClassMBeanInfo(StandardMBeanIntrospector.getInstance(), clazz, standardMBeanInterface);
    }
    
    private static <M> MBeanInfo getClassMBeanInfo(MBeanIntrospector<M> mBeanIntrospector, Class<?> clazz, Class<?> clazz2) throws NotCompliantMBeanException {
        return mBeanIntrospector.getClassMBeanInfo(clazz, mBeanIntrospector.getPerInterface(clazz2));
    }
    
    public static Class<?> getMBeanInterface(Class<?> clazz) {
        if (isDynamic(clazz)) {
            return null;
        }
        try {
            return getStandardMBeanInterface(clazz);
        }
        catch (NotCompliantMBeanException ex) {
            return null;
        }
    }
    
    public static <T> Class<? super T> getStandardMBeanInterface(Class<T> superclass) throws NotCompliantMBeanException {
        Class<Object> mBeanInterface = null;
        while (superclass != null) {
            mBeanInterface = findMBeanInterface(superclass, superclass.getName());
            if (mBeanInterface != null) {
                break;
            }
            superclass = superclass.getSuperclass();
        }
        if (mBeanInterface != null) {
            return mBeanInterface;
        }
        throw new NotCompliantMBeanException("Class " + superclass.getName() + " is not a JMX compliant Standard MBean");
    }
    
    public static <T> Class<? super T> getMXBeanInterface(Class<T> clazz) throws NotCompliantMBeanException {
        try {
            return MXBeanSupport.findMXBeanInterface(clazz);
        }
        catch (Exception ex) {
            throw throwException(clazz, ex);
        }
    }
    
    private static <T> Class<? super T> findMBeanInterface(Class<T> superclass, String s) {
        Class[] interfaces;
        int length;
        int i;
        Class<Object> implementsMBean;
        for (; superclass != null; superclass = superclass.getSuperclass()) {
            interfaces = superclass.getInterfaces();
            length = interfaces.length;
            i = 0;
            while (i < length) {
                implementsMBean = implementsMBean((Class)Util.cast(interfaces[i]), s);
                if (implementsMBean != null) {
                    return implementsMBean;
                }
                ++i;
            }
        }
        return null;
    }
    
    public static Descriptor descriptorForElement(AnnotatedElement annotatedElement) {
        if (annotatedElement == null) {
            return ImmutableDescriptor.EMPTY_DESCRIPTOR;
        }
        return descriptorForAnnotations(annotatedElement.getAnnotations());
    }
    
    public static Descriptor descriptorForAnnotations(Annotation[] array) {
        if (array.length == 0) {
            return ImmutableDescriptor.EMPTY_DESCRIPTOR;
        }
        HashMap<String, Object> fields = new HashMap();
        for (Annotation obj : array) {
            Class<Object> annotationType = obj.annotationType();
            Method[] methods = annotationType.getMethods();
            int n = 0;
            for (Method m : methods) {
                DescriptorKey descriptorKey = (DescriptorKey)m.getAnnotation(DescriptorKey.class);
                if (descriptorKey != null) {
                    String value = descriptorKey.value();
                    Object invoke;
                    try {
                        if (n == 0) {
                            ReflectUtil.checkPackageAccess(annotationType);
                            n = 1;
                        }
                        invoke = MethodUtil.invoke(m, obj, (Object[])null);
                    }
                    catch (RuntimeException ex) {
                        throw ex;
                    }
                    catch (Exception undeclaredThrowable) {
                        throw new UndeclaredThrowableException(undeclaredThrowable);
                    }
                    Object annotationToField = annotationToField(invoke);
                    Object put = fields.put(value, annotationToField);
                    if (put != null && !equals(put, annotationToField)) {
                        throw new IllegalArgumentException("Inconsistent values for descriptor field " + value + " from annotations: " + annotationToField + " :: " + put);
                    }
                }
            }
        }
        if (fields.isEmpty()) {
            return ImmutableDescriptor.EMPTY_DESCRIPTOR;
        }
        return new ImmutableDescriptor(fields);
    }
    
    static NotCompliantMBeanException throwException(Class<?> clazz, Throwable cause) throws NotCompliantMBeanException, SecurityException {
        if ((cause instanceof SecurityException)) {
            throw (SecurityException)cause;
        }
        if ((cause instanceof NotCompliantMBeanException)) {
            throw (NotCompliantMBeanException)cause;
        }
        NotCompliantMBeanException ex = new NotCompliantMBeanException(((clazz == null) ? "null class" : clazz.getName()) + ": " + ((cause == null) ? "Not compliant" : cause.getMessage()));
        ex.initCause(cause);
        throw ex;
    }
    
    private static Object annotationToField(Object o) {
        if (o == null) {
            return null;
        }
        if (o instanceof Number || o instanceof String || o instanceof Character || o instanceof Boolean || o instanceof String[]) {
            return o;
        }
        Class<Object> class1 = o.getClass();
        if (class1.isArray()) {
            if (class1.getComponentType().isPrimitive()) {
                return o;
            }
            Object[] array = (Object[])(Object[])o;
            String[] array2 = new String[array.length];
            for (int i = 0; i < array.length; ++i) {
                array2[i] = (String)annotationToField(array[i]);
            }
            return array2;
        }
        else {
            if ((o instanceof Class)) {
                return ((Class)o).getName();
            }
            if ((o instanceof Enum)) {
                return ((Enum)o).name();
            }
            if (Proxy.isProxyClass(class1)) {
                class1 = class1.getInterfaces()[0];
            }
            throw new IllegalArgumentException("Illegal type for annotation element using @DescriptorKey: " + class1.getName());
        }
    }
    
    private static boolean equals(Object o, Object o2) {
        return Arrays.deepEquals(new Object[] { o }, new Object[] { o2 });
    }
    
    private static <T> Class<? super T> implementsMBean(Class<T> clazz, String str) {
        String string = str + "MBean";
        if (clazz.getName().equals(string)) {
            return clazz;
        }
        Class[] interfaces = clazz.getInterfaces();
        int i = 0;
        while (i < interfaces.length) {
            if (interfaces[i].getName().equals(string) && (Modifier.isPublic(interfaces[i].getModifiers()) || Introspector.ALLOW_NONPUBLIC_MBEAN)) {
                return (Class)Util.cast(interfaces[i]);
            }
            ++i;
        }
        return null;
    }
    
    public static Object elementFromComplex(Object obj, String s) throws AttributeNotFoundException {
        try {
            if (obj.getClass().isArray() && s.equals("length")) {
                return Integer.valueOf(Array.getLength(obj));
            }
            if ((obj instanceof CompositeData)) {
                return ((CompositeData)obj).get(s);
            }
            Class<Object> class1 = obj.getClass();
            Method m = null;
            if (Introspector.BeansHelper.isAvailable()) {
                Object[] propertyDescriptors = Introspector.BeansHelper.getPropertyDescriptors(Introspector.BeansHelper.getBeanInfo(class1));
                int length = propertyDescriptors.length;
                int i = 0;
                while (i < length) {
                    Object o = propertyDescriptors[i];
                    if (Introspector.BeansHelper.getPropertyName(o).equals(s)) {
                        m = Introspector.BeansHelper.getReadMethod(o);
                        break;
                    }
                    ++i;
                }
            }
            else {
                m = Introspector.SimpleIntrospector.getReadMethod(class1, s);
            }
            if (m != null) {
                ReflectUtil.checkPackageAccess(m.getDeclaringClass());
                return MethodUtil.invoke(m, obj, new Class[0]);
            }
            throw new AttributeNotFoundException("Could not find the getter method for the property " + s + " using the Java Beans introspector");
        }
        catch (InvocationTargetException cause) {
            throw new IllegalArgumentException(cause);
        }
        catch (AttributeNotFoundException ex) {
            throw ex;
        }
        catch (Exception cause2) {
            throw (AttributeNotFoundException)EnvHelp.initCause(new AttributeNotFoundException(cause2.getMessage()), cause2);
        }
    }
    
    static {
        ALLOW_NONPUBLIC_MBEAN = Boolean.parseBoolean((String)AccessController.doPrivileged(new GetPropertyAction("jdk.jmx.mbeans.allowNonPublic")));
    }
    
    private static class SimpleIntrospector
    {
        private static final String GET_METHOD_PREFIX = "get";
        private static final String IS_METHOD_PREFIX = "is";
        private static final Map<Class<?>, SoftReference<List<Method>>> cache;
        
        private static List<Method> getCachedMethods(Class<?> clazz) {
            SoftReference<Object> softReference = (SoftReference)Introspector.SimpleIntrospector.cache.get(clazz);
            if (softReference != null) {
                List<Object> list = (List)softReference.get();
                if (list != null) {
                    return list;
                }
            }
            return null;
        }
        
        static boolean isReadMethod(Method method) {
            if (Modifier.isStatic(method.getModifiers())) {
                return false;
            }
            String name = method.getName();
            if (method.getParameterTypes().length == 0 && name.length() > 2) {
                if (name.startsWith("is")) {
                    return method.getReturnType() == Boolean.TYPE;
                }
                if (name.length() > 3 && name.startsWith("get")) {
                    return method.getReturnType() != Void.TYPE;
                }
            }
            return false;
        }
        
        static List<Method> getReadMethods(Class<?> clazz) {
            List<Method> cachedMethods = getCachedMethods(clazz);
            if (cachedMethods != null) {
                return cachedMethods;
            }
            List<Method> eliminateCovariantMethods = MBeanAnalyzer.eliminateCovariantMethods(StandardMBeanIntrospector.getInstance().getMethods(clazz));
            LinkedList referent = new LinkedList();
            for (Method method : eliminateCovariantMethods) {
                if (isReadMethod(method)) {
                    if (method.getName().startsWith("is")) {
                        referent.add(0, method);
                    }
                    else {
                        referent.add(method);
                    }
                }
                else {
                    continue;
                }
            }
            Introspector.SimpleIntrospector.cache.put(clazz, new SoftReference(referent));
            return referent;
        }
        
        static Method getReadMethod(Class<?> clazz, String string) {
            string = string.substring(0, 1).toUpperCase(Locale.ENGLISH) + string.substring(1);
            String string2 = "get" + string;
            String string3 = "is" + string;
            for (Method method : getReadMethods(clazz)) {
                String name = method.getName();
                if (name.equals(string3) || name.equals(string2)) {
                    return method;
                }
                continue;
            }
            return null;
        }
        
        static {
            cache = Collections.synchronizedMap(new WeakHashMap());
        }
    }
    
    private static class BeansHelper
    {
        private static final Class<?> introspectorClass;
        private static final Class<?> beanInfoClass;
        private static final Class<?> getPropertyDescriptorClass;
        private static final Method getBeanInfo;
        private static final Method getPropertyDescriptors;
        private static final Method getPropertyName;
        private static final Method getReadMethod;
        
        private static Class<?> getClass(String name) {
            try {
                return Class.forName(name, true, null);
            }
            catch (ClassNotFoundException ex) {
                return null;
            }
        }
        
        private static Method getMethod(Class<?> clazz, String name, Class<?>... parameterTypes) {
            if (clazz != null) {
                try {
                    return clazz.getMethod(name, parameterTypes);
                }
                catch (NoSuchMethodException detailMessage) {
                    throw new AssertionError(detailMessage);
                }
            }
            else {
                return null;
            }
        }
        
        static boolean isAvailable() {
            return Introspector.BeansHelper.introspectorClass != null;
        }
        
        static Object getBeanInfo(Class<?> clazz) throws Exception {
            try {
                return Introspector.BeansHelper.getBeanInfo.invoke(null, new Object[] { clazz });
            }
            catch (InvocationTargetException detailMessage) {
                Throwable cause = detailMessage.getCause();
                if ((cause instanceof Exception)) {
                    throw (Exception)cause;
                }
                throw new AssertionError(detailMessage);
            }
            catch (IllegalAccessException detailMessage2) {
                throw new AssertionError(detailMessage2);
            }
        }
        
        static Object[] getPropertyDescriptors(Object obj) {
            try {
                return (Object[])(Object[])Introspector.BeansHelper.getPropertyDescriptors.invoke(obj, new Object[0]);
            }
            catch (InvocationTargetException detailMessage) {
                Throwable cause = detailMessage.getCause();
                if ((cause instanceof RuntimeException)) {
                    throw (RuntimeException)cause;
                }
                throw new AssertionError(detailMessage);
            }
            catch (IllegalAccessException detailMessage2) {
                throw new AssertionError(detailMessage2);
            }
        }
        
        static String getPropertyName(Object obj) {
            try {
                return (String)Introspector.BeansHelper.getPropertyName.invoke(obj, new Object[0]);
            }
            catch (InvocationTargetException detailMessage) {
                Throwable cause = detailMessage.getCause();
                if ((cause instanceof RuntimeException)) {
                    throw (RuntimeException)cause;
                }
                throw new AssertionError(detailMessage);
            }
            catch (IllegalAccessException detailMessage2) {
                throw new AssertionError(detailMessage2);
            }
        }
        
        static Method getReadMethod(Object obj) {
            try {
                return (Method)Introspector.BeansHelper.getReadMethod.invoke(obj, new Object[0]);
            }
            catch (InvocationTargetException detailMessage) {
                Throwable cause = detailMessage.getCause();
                if ((cause instanceof RuntimeException)) {
                    throw (RuntimeException)cause;
                }
                throw new AssertionError(detailMessage);
            }
            catch (IllegalAccessException detailMessage2) {
                throw new AssertionError(detailMessage2);
            }
        }
        
        static {
            introspectorClass = getClass("java.beans.Introspector");
            beanInfoClass = ((Introspector.BeansHelper.introspectorClass == null) ? null : getClass("java.beans.BeanInfo"));
            getPropertyDescriptorClass = ((Introspector.BeansHelper.beanInfoClass == null) ? null : getClass("java.beans.PropertyDescriptor"));
            getBeanInfo = getMethod(Introspector.BeansHelper.introspectorClass, "getBeanInfo", new Class[] { Class.class });
            getPropertyDescriptors = getMethod(Introspector.BeansHelper.beanInfoClass, "getPropertyDescriptors", new Class[0]);
            getPropertyName = getMethod(Introspector.BeansHelper.getPropertyDescriptorClass, "getName", new Class[0]);
            getReadMethod = getMethod(Introspector.BeansHelper.getPropertyDescriptorClass, "getReadMethod", new Class[0]);
        }
    }
}
